@@ -2,14 +2,14 @@
 #'
 #' Trains a distributional random forest that can be used to estimate
 #' statistical functional F(P(Y | X)) for possibly multivariate response Y.
-#'
 #' @param X The covariates used in the regression. Can be either a matrix of numerical values, or a data.frame with characters and factors. In the latter case,
 #'   one-hot-encoding will be implicitely used.
 #' @param Y The (multivariate) outcome. A matrix or data.frame of numeric values.
 #' @param num.trees Number of trees grown in the forest. Default is 500.
 #' @param splitting.rule a character value. The type of splitting rule used, can be either "CART" or "FourierMMD".
 #' @param num.features a numeric value, in case of "FourierMMD", the number of random features to sample.
-#' @param bandwidth a numeric value, the bandwidth of the Gaussian kernel used in case of "FourierMMD".
+#' @param bandwidth a numeric value, the bandwidth of the Gaussian kernel used in case of "FourierMMD", by default the value is NULL and the median heuristic is used.
+#' @param response.scaling a boolean value, should the reponses be globally scaled at first.
 #' @param node.scaling a boolean value, should the responses be scaled or not by node.
 #' @param sample.weights (experimental) Weights given to an observation in estimation.
 #'                       If NULL, each observation is given the same weight. Default is NULL.
@@ -53,6 +53,7 @@
 #' @param num.threads Number of threads used in training. By default, the number of threads is set
 #'                    to the maximum hardware concurrency.
 #' @param seed The seed of the C++ random number generator.
+#' @param compute.variable.importance boolean, should the variable importance be computed in the object.
 #'
 #' @return A trained distributional random forest object.
 #'
@@ -71,9 +72,9 @@
 #'
 #' # Predict on out-of-bag training samples.
 #' cor.oob.pred <- predict(drf.forest,  functional = "cor")
-#' 
+#'
 #' # Train a distributional random forest with "FourierMMD" splitting rule.
-#' n <- 50
+#' n <- 100
 #' p <- 2
 #' X <- matrix(rnorm(n * p), n, p)
 #' Y <- X + matrix(rnorm(n * p), ncol=p)
@@ -95,7 +96,8 @@ drf <-               function(X, Y,
                               num.trees = 500,
                               splitting.rule = "FourierMMD",
                               num.features = 10,
-                              bandwidth = 1.0,
+                              bandwidth = NULL,
+                              response.scaling = TRUE,
                               node.scaling = FALSE,
                               sample.weights = NULL,
                               clusters = NULL,
@@ -111,24 +113,24 @@ drf <-               function(X, Y,
                               ci.group.size = 2,
                               compute.oob.predictions = TRUE,
                               num.threads = NULL,
-                              seed = stats::runif(1, 0, .Machine$integer.max)) {
-  
+                              seed = stats::runif(1, 0, .Machine$integer.max),
+                              compute.variable.importance = FALSE) {
+
   # initial checks for X and Y
   if (is.data.frame(X)) {
-    
+
     if (is.null(names(X))) {
-      stop("teh regressor should be named if provided under data.frame format.")
+      stop("the regressor should be named if provided under data.frame format.")
     }
-    
+
     if (any(apply(X, 2, class) %in% c("factor", "character"))) {
-      
       any.factor.or.character <- TRUE
       X.mat <- as.matrix(fastDummies::dummy_cols(X, remove_selected_columns = TRUE))
     } else {
-      
       any.factor.or.character <- FALSE
       X.mat <- as.matrix(X)
     }
+
     mat.col.names.df <- names(X)
     mat.col.names <- colnames(X.mat)
   } else {
@@ -137,16 +139,20 @@ drf <-               function(X, Y,
     mat.col.names.df <- NULL
     any.factor.or.character <- FALSE
   }
-  
+
   if (is.data.frame(Y)) {
-    
+
     if (any(apply(Y, 2, class) %in% c("factor", "character"))) {
       stop("Y should only contain numeric variables.")
     }
     Y <- as.matrix(Y)
   }
   
-  
+  if (is.vector(Y)) {
+    Y <- matrix(Y,ncol=1)
+  }
+
+
   validate_X(X.mat)
   validate_sample_weights(sample.weights, X.mat)
   #Y <- validate_observations(Y, X)
@@ -156,8 +162,21 @@ drf <-               function(X, Y,
 
   all.tunable.params <- c("sample.fraction", "mtry", "min.node.size", "honesty.fraction",
                           "honesty.prune.leaves", "alpha", "imbalance.penalty")
+  
+  # should we scale or not the data
+  if (response.scaling) {
+    Y.transformed <- scale(Y)
+  } else {
+    Y.transformed <- Y
+  }
 
-  data <- create_data_matrices(X.mat, outcome = scale(Y), sample.weights = sample.weights)
+  data <- create_data_matrices(X.mat, outcome = Y.transformed, sample.weights = sample.weights)
+  
+  # bandwidth using median heuristic by default
+  if (is.null(bandwidth)) {
+    bandwidth <- medianHeuristic(Y.transformed)
+  }
+  
 
   args <- list(num.trees = num.trees,
                clusters = clusters,
@@ -174,7 +193,7 @@ drf <-               function(X, Y,
                compute.oob.predictions = compute.oob.predictions,
                num.threads = num.threads,
                seed = seed,
-               num_features = num.features, 
+               num_features = num.features,
                bandwidth = bandwidth,
                node_scaling = ifelse(node.scaling, 1, 0))
 
@@ -187,7 +206,7 @@ drf <-               function(X, Y,
    } else {
      stop("splitting rule not available.")
    }
-   
+
    class(forest) <- c("drf")
    forest[["ci.group.size"]] <- ci.group.size
    forest[["X.orig"]] <- X.mat
@@ -201,6 +220,10 @@ drf <-               function(X, Y,
    forest[["mat.col.names.df"]] <- mat.col.names.df
    forest[["any.factor.or.character"]] <- any.factor.or.character
    
+   if (compute.variable.importance) {
+     forest[['variable.importance']] <- variableImportance(forest, h = bandwidth)
+   }
+
    forest
 }
 
@@ -215,31 +238,34 @@ drf <-               function(X, Y,
 #'                matrix, and that the columns must appear in the same order.
 #' @param functional which type of statistical functional. One option between:
 #' \itemize{
-#'  \item{"mean"}{the conditional mean, the returned value is a list containing a matrix \code{mean} of size \code{n} x \code{f}, 
+#'  \item{"mean"}{the conditional mean, the returned value is a list containing a matrix \code{mean} of size \code{n} x \code{f},
 #'  where \code{n} denotes the number of observation in \code{newdata} and \code{f} the dimension of the \code{transformation}.}
-#'  \item{"sd"}{the conditional standard deviation, the returned value is a list containing a matrix \code{sd} of size \code{n} x \code{f}, 
+#'  \item{"sd"}{the conditional standard deviation, the returned value is a list containing a matrix \code{sd} of size \code{n} x \code{f},
 #'  where \code{n} denotes the number of observation in \code{newdata} and \code{f} the dimension of the \code{transformation}.}
-#'  \item{"quantile"}{the conditional quantiles, the returned value is a list containing an array \code{quantile} of size \code{n} x \code{f}  x \code{q}, 
+#'  \item{"quantile"}{the conditional quantiles, the returned value is a list containing an array \code{quantile} of size \code{n} x \code{f}  x \code{q},
 #'  where \code{n} denotes the number of observation in \code{newdata}, \code{f} the dimension of the \code{transformation} and \code{q} the number of desired quantiles.}
-#'  \item{"cor"}{the conditional correlation, the returned value is a list containing an array \code{cor} of size \code{n} x \code{f}  x \code{f}, 
+#'  \item{"cor"}{the conditional correlation, the returned value is a list containing an array \code{cor} of size \code{n} x \code{f}  x \code{f},
 #'  where \code{n} denotes the number of observation in \code{newdata}, \code{f} the dimension of the \code{transformation}.}
-#' \item{"cov"}{the conditional covariance, the returned value is a list containing an array \code{cor} of size \code{n} x \code{f}  x \code{f}, 
+#' \item{"cov"}{the conditional covariance, the returned value is a list containing an array \code{cor} of size \code{n} x \code{f}  x \code{f},
 #'  where \code{n} denotes the number of observation in \code{newdata}, \code{f} the dimension of the \code{transformation}.}
-#'  \item{"cdf"}{the conditional cumulative distribution function, the returned value is a list containing a list of functions \code{cdf} of size \code{n}, 
-#'  where \code{n} denotes the number of observation in \code{newdata}.  Here the transformation should be uni-dimensional.} 
-#'  \item{"normalPredictionScore"}{a prediction score based on an asymptotic normality assumption, the returned value is a list containing a list of functions \code{normalPredictionScore} of size \code{n}, 
+#'  \item{"normalPredictionScore"}{a prediction score based on an asymptotic normality assumption, the returned value is a list containing a list of functions \code{normalPredictionScore} of size \code{n},
 #'  where \code{n} denotes the number of observation in \code{newdata}. Here the transformation should be uni-dimensional.}
+#'  \item{"custom"}{a custom function provided by the user, the returned value is a list containing a matrix \code{custom} of size \code{n} x \code{f},
+#'  where \code{n} denotes the number of observation in \code{newdata} and \code{f} the dimension of the output of the function \code{custom.functional}.}
+#'  \item{"MQ"}{multivariate quantiles, return a list containing a matrix of the inputed ranks u (that should be provided as an argument of the predict function) along with a list of the different corresponding MQ (same size as u).}
 #' }
 #' @param transformation a function giving a transformation of the responses, by default if NULL, the identity \code{function(y) y} is used.
 #' @param num.threads Number of threads used in training. If set to NULL, the software
 #'                    automatically selects an appropriate amount.
+#' @param custom.functional a function giving the custom functional when \code{functional} is set to "custom". This should be a function \code{f(y,w)} using the
+#'   training response matrix \code{y} and the weights \code{w} at a single testing point.
 #' @param ... additional parameters.
 #'
-#' @return a list containing an entry with the same name as the functional selected. 
+#' @return a list containing an entry with the same name as the functional selected.
 #'
 #' @examples
 #' # Train a distributional random forest with CART splitting rule.
-#' n <- 50
+#' n <- 100
 #' p <- 2
 #' X <- matrix(rnorm(n * p), n, p)
 #' Y <- X + matrix(rnorm(n * p), ncol=p)
@@ -252,7 +278,7 @@ drf <-               function(X, Y,
 #'
 #' # Predict on out-of-bag training samples.
 #' cor.oob.pred <- predict(drf.forest,  functional = "cor")
-#' 
+#'
 #' # Train a distributional random forest with "FourierMMD" splitting rule.
 #' n <- 100
 #' p <- 2
@@ -271,259 +297,284 @@ drf <-               function(X, Y,
 #'
 #' @method predict drf
 #' @export
-#' 
-predict.drf <- function(object, 
+#'
+predict.drf <- function(object,
                         newdata = NULL,
                         transformation = NULL,
                         functional = NULL,
                         num.threads = NULL,
+                        custom.functional = function(y, w) apply(y,2,sum(y*w)),
                         ...) {
-  
+
   # if the newdata is a data.frame we should be careful about the non existing levels
   if (!is.null(newdata) && is.data.frame(newdata)) {
-    
-    
+
+
     if (is.data.frame(newdata) && !object$is.df.X) {
       stop("data.frame for newdata is accepted only if it was used for training data.")
     }
     if (ncol(newdata) != length(object$mat.col.names.df)) {
       stop("newdata should have the same dimension as the training data.")
     }
-    
+
     names(newdata) <- object$mat.col.names.df
-    
+
     # check if factor or not
     if (!object$any.factor.or.character) {
       newdata.mat <- as.matrix(newdata)
     } else {
-      newdata.mat <- as.matrix(fastDummies::dummy_cols(.data = newdata, 
+      newdata.mat <- as.matrix(fastDummies::dummy_cols(.data = newdata,
                                                        remove_selected_columns = TRUE))
-    
-    
+
+
       # define the modifications of the columns to do
       col.to.remove <- setdiff(colnames(newdata.mat), object$mat.col.names)
       col.to.add <- setdiff(object$mat.col.names, colnames(newdata.mat))
-      
+
       # col to remove
       newdata.mat <- newdata.mat[,!(colnames(newdata.mat)%in%col.to.remove), drop = FALSE]
-  
+
       # col to add
       prev.nb.col <- ncol(newdata.mat)
       prev.col.names <- colnames(newdata.mat)
-      
+
       for (col in col.to.add) {
         newdata.mat <- cbind(newdata.mat, 0)
       }
-    
+
       colnames(newdata.mat) <- c(prev.col.names, col.to.add)
-    
+
       newdata.mat <- newdata.mat[,object$mat.col.names]
     }
   } else if (!is.null(newdata)) {
     newdata.mat <- newdata
   }
-  
-  
-  # support vector as input 
+
+
+  # support vector as input
   if (!is.null(newdata) && is.null(dim(newdata.mat))) {
     newdata.mat <- matrix(newdata.mat, 1)
   } else if (is.null(newdata)) {
     newdata.mat <- NULL
   }
-  
+
   # get the weights which are used in a second step
-  w <- get_sample_weights(forest = object, 
-                          newdata = newdata.mat, 
+  w <- get_sample_weights(forest = object,
+                          newdata = newdata.mat,
                           num.threads = num.threads)
-  
-  
-  if (!is.null(transformation) && !(functional %in% c("mean", "quantile", "sd", 
+
+
+  if (!is.null(transformation) && !(functional %in% c("mean", "quantile", "sd",
                                                       "cor", "cov",
                                                       "normalPredictionScore", "cdf"))) {
     stop("transformation not available.")
   }
-  
-  
+
+
   if (is.null(transformation)) {
     transformation <- function(y) y
   }
-  
+
   if (is.null(functional)) {
-    
-    # return the weights 
-    return(list(weights = w, 
+
+    # return the weights
+    return(list(weights = w,
                 y = object$Y.orig))
-    
+
   } else if (functional %in% c("mean",
-                               "quantile", 
+                               "quantile",
                                "sd")) {
-    
-    
+
+
     # get the additional parameters
     add.param <- list(...)
-    
+
     if (functional == "quantile" && is.null(add.param$quantiles)) {
       stop("additional parameter quantiles should be provided when functional is quantile.")
     }
-    
+
     # compute the functional on the training set
-    functional.t <- t(apply(object$Y.orig, 
-                            1, 
+    functional.t <- t(apply(object$Y.orig,
+                            1,
                             function(yy) transformation(yy)))
-    
+
     # check length one (R size management)
     if (length(transformation(object$Y.orig[1,])) == 1) {
       functional.t <- t(functional.t)
     }
-      
+
       # in case of quantile regression
       if (!is.null(add.param$quantiles)) {
-        
-        functional.val <- lapply(1:ncol(functional.t), function(j) t(apply(w, 1, function(ww) spatstat::weighted.quantile(x = functional.t[ww!=0, j], 
-                                                                                                                          w = ww[ww!=0], 
+
+        functional.val <- lapply(1:ncol(functional.t), function(j) t(apply(w, 1, function(ww) weighted.quantile(x = functional.t[ww!=0, j],
+                                                                                                                          w = ww[ww!=0],
                                                                                                                           probs = add.param$quantiles))))
-        quantile.array <- array(dim = c(nrow(w), ncol(functional.t), length(add.param$quantiles)), 
-                                dimnames = list(NULL, NULL, paste("q=", round(add.param$quantiles, 2), sep="")))                                                                      
-      
+        quantile.array <- array(dim = c(nrow(w), ncol(functional.t), length(add.param$quantiles)),
+                                dimnames = list(NULL, NULL, paste("q=", round(add.param$quantiles, 2), sep="")))
+
         for (i in 1:length(functional.val)) {
-        
+
         if (length(add.param$quantile) == 1) {
           functional.val[[i]] <- t(functional.val[[i]])
         }
         #colnames(functional.val[[i]]) <- paste("q=", round(add.param$quantiles, 2), sep="")
         quantile.array[,i,] <- functional.val[[i]]
       }
-        
-        
+
+
         return(list(quantile = quantile.array))
-        
+
       }
     }
-    
-    # if no quantiles provided then conditional mean or sd, possibly multi-dimensional
-    if (functional == "mean") {
-      
+
+    if (functional == "custom") {
+
+      if (!is.null(transformation)) {
+        stop("when custom functional is called, transformation should be the identity.")
+      }
+
+      custom <- t(apply(w, 1, function(ww) custom.functional(object$Y.orig, ww)))
+
+      return(list(custom = custom))
+
+    } else if (functional == "mean") {
+
       functional.mean <- t(apply(w, 1, function(ww) ww%*%functional.t))
-      
+
       # check length one (R size management)
       if (length(transformation(object$Y.orig[1,])) == 1) {
-        
+
         functional.mean <- t(functional.mean)
       }
-      
+
       #colnames(functional.mean) <- colnames(object$Y.orig)
-        
+
       return(list(mean = functional.mean))
-      
+
     } else if (functional == "sd") {
-      
+
       functional.mean <- t(apply(w, 1, function(ww) ww%*%functional.t))
       functional.mean2 <- t(apply(w, 1, function(ww) ww%*%(functional.t)^2))
-      
+
       # check length one (R size management)
       if (length(transformation(object$Y.orig[1,])) == 1) {
-        
-        
+
+
         functional.mean <- t(functional.mean)
         functional.mean2 <- t(functional.mean2)
       }
-      
+
       functional.sd <- sqrt(functional.mean2-(functional.mean)^2)
-      #colnames(functional.sd) <- colnames(object$Y.orig)      
-      
+      #colnames(functional.sd) <- colnames(object$Y.orig)
+
       return(list(sd = functional.sd))
-    
+
   } else if (functional == "cor") {
-    
+
     # compute the functional on the training set
-    functional.t <- t(apply(object$Y.orig, 
-                            1, 
+    functional.t <- t(apply(object$Y.orig,
+                            1,
                             function(yy) transformation(yy)))
-    
+
     # check length one (R size management)
     if (length(transformation(object$Y.orig[1,])) == 1) {
       stop("cor available only for multi-dimensional transformation.")
     }
-    
+
     cor.mat <- array(1, dim = c(nrow(w), ncol(functional.t), ncol(functional.t)),
                      dimnames = list(NULL, NULL, NULL))
-    
+
     for (i in 1:nrow(w)) {
       cor.mat[i,,] <- stats::cov.wt(x = functional.t, wt = as.numeric(w[i,]), cor = TRUE)$cor
-    } 
-    
+    }
+
     return(list(cor = cor.mat))
-    
+
   } else if (functional == "cov") {
-    
+
     # compute the functional on the training set
-    functional.t <- t(apply(object$Y.orig, 
-                            1, 
+    functional.t <- t(apply(object$Y.orig,
+                            1,
                             function(yy) transformation(yy)))
-    
+
     # check length one (R size management)
     if (length(transformation(object$Y.orig[1,])) == 1) {
       stop("cor available only for multi-dimensional transformation.")
     }
-    
+
     cov.mat <- array(1, dim = c(nrow(w), ncol(functional.t), ncol(functional.t)),
                      dimnames = list(NULL, NULL, NULL))
-    
+
     for (i in 1:nrow(w)) {
       cov.mat[i,,] <- stats::cov.wt(x = functional.t, wt = as.numeric(w[i,]))$cov
     }
-    
+
     return(list(cov = cov.mat))
-    
+
   }  else if (functional == "normalPredictionScore") {
-    
+
     # compute the functional on the training set
-    functional.t <- t(apply(object$Y.orig, 
-                            1, 
+    functional.t <- t(apply(object$Y.orig,
+                            1,
                             function(yy) transformation(yy)))
-    
+
     # check length one (R size management)
     if (length(transformation(object$Y.orig[1,])) == 1) {
       stop("cor available only for multi-dimensional transformation.")
     }
-    
+
     means <- t(apply(w, 1, function(ww) ww%*%functional.t))
-    
+
     covs <- array(1, dim = c(nrow(w), ncol(functional.t), ncol(functional.t)))
-    
+
     for (i in 1:nrow(w)) {
       covs[i,,] <- stats::cov.wt(x = functional.t, wt = as.numeric(w[i,]))$cov
-    } 
-    
+    }
+
     # dims
     n <- nrow(object$Y.orig)
     d <- ncol(object$Y.orig)
-    
+
     funs <- lapply(1:nrow(w), function(i) {
                     inv.cov <- solve(covs[i,,])
-                    
+
                     return(function(y) (n/(n+1))*((n-d)/(d*(n-1)))*as.numeric((y-means[i,])%*%inv.cov%*%(y-means[i,])))
                   })
-    
+
     return(list(normalPredictionScore = funs))
-       
-  } else if (functional == "cdf") {
+  } else if (functional == "MQ") {
+
+    # compute the functional on the training set
+    #functional.t <- t(apply(object$Y.orig,
+    #                        1,
+    #                        function(yy) transformation(yy)))
+
+    u <- list(...)$u
     
-    functional.t <- apply(object$Y.orig, 
-                            1, 
-                            function(yy) transformation(yy))
-    
-    # check length one (R size management)
-    if (length(transformation(object$Y.ori[1,])) != 1) {
-      stop("multi-dimensional ecdf not available.")
-    } else {
-      functional.t <- t(functional.t)
+    if (!is.matrix(u) || ncol(u)!=ncol(object$Y.orig)) {
+      stop("imcompatible u with the response y.")
     }
-    
-    funs <- lapply(1:nrow(w), function(i) {
-      return(function(y) spatstat::ewcdf(x = functional.t, weights = as.numeric(w[i,]))(y))
-    })
-    
-    return(list(cdf = funs))
-  } 
+
+    # compute the cost between the provided u's and the y's
+    costm <- t(apply(object$Y.orig, 1, function(yy) apply(u, 1, function(uu) {
+      sum((yy-uu)^2)
+    })))
+
+    # get the transport solution
+    info.mq <- apply(w,
+                      1,
+                      function(ww) {
+                        ids.in <- which(ww!=0)
+                        tr <- transport::transport(costm[ids.in,], a = ww[ids.in], b = rep(1/nrow(u),nrow(u)), fullreturn = TRUE)
+                        ids.y <- apply(tr$primal, 2, function(x) sample(1:length(x), size = 1, replace = FALSE, prob = x))
+                        return(list(ids.y=ids.y, ids.in=ids.in))
+                     })
+
+    # get one version of the multimap
+    yhat <- lapply(info.mq, function(info) {object$Y.orig[info$ids.in,,drop=F][info$ids.y,,drop=F]})
+    return(list(multvariateQuantiles = list(yhat = yhat, u = u)))
+
+  } else {
+    stop("functional not implemented!")
+  }
 }
